from datetime import timedelta
from movva_tools.models.campaign_models import RapidProCampaignEvents
from movva_tools.serializers.base_serializer import (
    BaseSerializer, serializer_fields, date_field_format
)


class CampaingEventSerializer(BaseSerializer):
    class Meta:
        model = RapidProCampaignEvents

    def __init__(self, *args, config_time, **kwargs):
        super().__init__(*args, **kwargs)
        self.config_time = config_time

    uuid = serializer_fields.UUID(dump_only=True)
    delivery_hour = serializer_fields.Integer(dump_only=True)
    offset = serializer_fields.Integer(dump_only=True)
    unit = serializer_fields.String(dump_only=True)
    created_on = date_field_format
    config_time = serializer_fields.Method("format_reference_date", dump_only=True)

    def format_reference_date(self, obj):
        if self.config_time['delivery_hour'] == -1:
            return self.config_time['delivery_time'].strftime("%d/%m/%Y %H:%M")
        else:
            return (
                self.config_time['delivery_time'] + timedelta(
                    hours=self.config_time['delivery_hour']
                )
            ).strftime("%d/%m/%Y %H:%M")
        # Aqui, você formata a data conforme necessário
        # Estou assumindo que `self.reference_date` já está em um formato de data/hora válido



from movva_tools.services.api_token_service import APITokenService
from movva_tools.services.base_service import BaseService
from movva_tools.services.campaign_service import CampaingService
from movva_tools.services.channel_service import ChannelService
from movva_tools.services.contacts_service import ContactService
from movva_tools.services.flow_service import (
    FlowService, FlowLabelService, CopyFlowAndScheduleService,
    FlowDependenciesService
)
from movva_tools.services.organization_service import OrganizationService
from movva_tools.services.user_service import UserService
from movva_tools.services.messages_service import MessageService



import uuid

from movva_tools.databases import Base

from sqlalchemy import (
    JSON, UUID, Boolean, Column, DateTime,
    Integer, String,
)


class RapidProOrganization(Base):
    __tablename__ = 'orgs_org'

    id = Column(Integer, primary_key=True)
    is_active = Column(Boolean)
    created_on = Column(DateTime)
    modified_on = Column(DateTime)
    uuid = Column(UUID(as_uuid=True), default=uuid.uuid4)
    name = Column(String)
    plan = Column(String)
    stripe_customer = Column(String)
    language = Column(String)
    timezone = Column(String)
    date_format = Column(String)
    config = Column(String)
    slug = Column(String)
    is_anon = Column(Boolean)
    brand = Column(String)
    surveyor_password = Column(String)
    country_id = Column(Integer)
    created_by_id = Column(Integer)
    modified_by_id = Column(Integer)
    parent_id = Column(Integer)
    is_flagged = Column(Boolean)
    is_suspended = Column(Boolean)
    uses_topups = Column(Boolean)
    is_multi_org = Column(Boolean)
    is_multi_user = Column(Boolean)
    plan_end = Column(DateTime)
    plan_start = Column(DateTime)
    limits = Column(JSON)
    released_on = Column(DateTime)
    deleted_on = Column(DateTime)
    flow_languages = Column(String)
    api_rates = Column(JSON)

    def __repr__(self):
        return f'<{self.uuid}:{self.name}> slug:{self.slug}'



from datetime import datetime
import pandas as pd
from more_itertools import chunked
from typing import Dict, List, Union
import json

DATE_FORMAT = "%d/%m/%Y"
TIME_FORMAT = "%H:%M:%S"


def set_prenome_column(df: pd.DataFrame) -> pd.DataFrame:
    try:
        df['PRENOME'] = df['NOME'].apply(lambda x: x.split()[0])
        return df
    except KeyError:
        raise KeyError('A chave NOME é obrigatória na planilha.')


def date_formatter(df: pd.DataFrame, birth_date_column: str) -> pd.DataFrame:
    df[f'{birth_date_column}_FORMATADA'] = pd.to_datetime(
        df[birth_date_column], format='"%Y-%m-%dT%H:%M:%SZ"'
    )

    df[f'{birth_date_column}_FORMATADA'] = df[f'{birth_date_column}_FORMATADA'].astype(str)

    return df


def strip_whitespaces(data: str):
    return data.strip()


def replace_newline_character_for_space(data: str):
    return data.replace('\n', ' ')


def break_list_into_groups(input_list, chunk_size=100) -> list:
    """
        Determina uma lista que contém lotes de tamanho informado em
        chunk_size.
    """
    return list(chunked(input_list, chunk_size))


def datetime_string_parser(date_str: str, pattern: str):
    if not date_str:
        return date_str
    try:
        data = datetime.strptime(date_str, pattern)
        return data
    except ValueError:
        raise ValueError('Incorrect format of date and time.')


def json_representation(
    keys: list, serialized_data: List[Dict]
) -> Union[List[Dict], Dict]:

    result = {key: json.loads(value) for key, value in zip(keys, serialized_data)}
    return result


def serialize(serializer_class, instance, config_time=None):
    if config_time:
        serializer = serializer_class(config_time=config_time)
    else:
        serializer = serializer_class()
    return serializer.dumps(instance)



from temba_client.v2 import TembaClient
import asyncio

from datetime import datetime
from pandas import DataFrame
from movva_tools.utils import set_prenome_column
from more_itertools import chunked


class AsyncTembaClient(TembaClient):
    """
        Classe modificada do Client Python do RapidPro para criar requisições
        assíncronas aos recursos da API do RapidPro.
    """

    async def create_contact_async(
        self, name=None, language=None, urns=None, fields=None, groups=None
    ):
        """
        Asynchronously creates a new contact
        :param str name: full name
        :param str language: the language code, e.g. "eng"
        :param list[str] urns: list of URN strings
        :param dict[str,str] fields: dictionary of contact field values
        :param list groups: list of group objects, UUIDs or names
        :return: the new contact
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self.create_contact, name, language, urns, fields, groups)


class ImportContacts:
    """
        Determina métodos para cadastrar contatos no RapidPro.
    """

    def __init__(self, url, organization_token) -> None:
        self.__url = url
        self.__organization_token = organization_token
        self.client = None
        self.contact_fields = []
        self.contact_labels = []
        self.pos_contacts_with_erros = []
        self.pos_contacts_with_success = []

    @staticmethod
    def break_list_into_groups(input_list, chunk_size=100) -> list:
        """
            Determina uma lista que contém lotes de tamanho informado em
            chunk_size.
        """
        return list(chunked(input_list, chunk_size))

    def set_client(self):
        """
            Inicia conexão com Client do RapidPro.
        """
        self.client = AsyncTembaClient(self.__url, self.__organization_token)
        return self.client

    def set_pattern_of_fields(self, fields: list):

        for index, field in enumerate(fields):
            fields[index] = field.upper()

        return fields

    def get_contact_fields_info(self) -> list:

        keys = self.client.get_fields().iterfetches(retry_on_rate_exceed=True)

        for register in keys:
            for f in register:
                self.contact_fields.append(f.key)
                self.contact_labels.append(f.key.upper())

    def insert_ddi_in_urns(self, numero):
        numero_str = str(numero)
        if numero_str.startswith('55'):
            return numero
        else:
            return '55' + numero_str

    def get_contact_fields_in_dataframe(self, dataframe: DataFrame):

        dataframe.rename(columns=lambda x: x.upper(), inplace=True)
        dataframe_fields = dataframe.columns

        intersection = set(self.contact_labels).intersection(dataframe_fields)

        df = dataframe.loc[:, list(intersection)]

        return df

    def create_group_contacts(self, suggested_group_name=None) -> str:
        """
            Cria um grupo para armazenar os contatos no padrão:
            imported_${dd/mm/aaaa} com data atual ou informado através
            do parâmetro suggested_group_name.
        """

        date = datetime.now().strftime("%d/%m/%Y")
        group_name = f'imported {date}'

        if suggested_group_name:
            group_name = suggested_group_name

        try:
            self.client.create_group(name=group_name)
        except Exception:
            return group_name

        return group_name

    def adjust_contact_fields_sheet(self, dataframe: DataFrame) -> list:

        dataframe.rename(columns=lambda x: x.lower(), inplace=True)

        return dataframe

    def set_tel_urn(self, urn: str) -> str:
        urn_pattern = 'tel:+'

        return f'{urn_pattern}{urn}'

    def set_whatsapp_urn(self, urn: str) -> str:
        urn_pattern = 'whatsapp:'

        return f'{urn_pattern}{urn}'

    def export_unsigned_contacts(self, df: DataFrame, export=True):
        """
            Filtra os contatos não cadastrados do DataFrame de contatos
            e com export=True, cria um arquivo excel com os dados dos contatos
            não cadastrados.
        """
        contatos_nao_cadastrados = df.iloc[self.pos_contacts_with_erros]

        if export:
            contatos_nao_cadastrados.to_excel('nao_cadastrados.xlsx')

        return contatos_nao_cadastrados

    def process_contacts(self, df: DataFrame, group_name=None, prenome_given=True) -> (DataFrame, str):
        """
            Método que predefine o fluxo de cadastro de contatos a partir de um dataframe.
        """

        self.set_client()

        self.get_contact_fields_info()

        df = self.get_contact_fields_in_dataframe(dataframe=df)

        if prenome_given:
            df = set_prenome_column(df)

        self.adjust_contact_fields_sheet(df)

        group = self.create_group_contacts(suggested_group_name=group_name)

        return (df, group)

    async def create_contact_async(self, payload, group, df, index):
        """
            Cadastra os contatos no Rapidpro e verifica se houve alguma
            falha na requisição de cadastro.
        """

        try:
            # Utiliza o método assíncrono da classe AsyncTembaClient
            response = await self.client.create_contact_async(
                name=payload['nome'],
                urns=[
                    self.set_tel_urn(payload['celular']),
                    self.set_whatsapp_urn(payload['celular'])
                ],
                fields=payload,
                groups=[group]
            )

            # Verificar se o cadastro foi bem-sucedido
            if not hasattr(response, 'uuid') and index not in self.pos_contacts_with_success:
                # Armazena a posição do contato que falhou no cadastro.
                self.pos_contacts_with_erros.append(index)
            else:
                self.pos_contacts_with_success.append(index)

        except Exception as e:
            # Armazena a posição do contato que falhou no cadastro.
            print(f"Erro ao cadastrar contato n° {index}: {e}")
            self.pos_contacts_with_erros.append(index)

        return self.pos_contacts_with_erros

    async def create_contacts_in_batches(
        self, payloads, group, batch_size=100
    ):
        """
            Cria lotes de contatos para evitar gargalos e facilitar rotina assíncrona.
        """
        # Divide a lista de payloads em lotes menores
        batches = [payloads[i:i + batch_size] for i in range(0, len(payloads), batch_size)]

        results = []

        for batch in batches:
            tasks = [self.create_contact_async(payload, group, batch, index) for index, payload in enumerate(batch)]
            # Executa as tarefas de criação em lote de forma assíncrona
            batch_results = await asyncio.gather(*tasks)
            results.extend(batch_results)

        return results

    async def import_contacts(self, df: DataFrame, group_name=None, urn_column=None):

        """
            Cadastra os contatos de um DataFrame no RapidPro.
        """

        df, group = self.process_contacts(df, group_name)

        df = df.fillna('')

        df['celular'] = df['celular'].apply(self.insert_ddi_in_urns)

        # Cria uma lista de payloads com os dados de cada contato
        payloads = [row.to_dict() for _, row in df.iterrows()]

        # Executa a criação de contatos em lotes de forma assíncrona
        await self.create_contacts_in_batches(payloads, group)

        nao_cadastrados = self.export_unsigned_contacts(df, export=False)

        if not nao_cadastrados.empty:
            print('Alguns contatos não foram cadastrados com sucesso. Um arquivo será gerado com os contatos não cadastrados...')
        else:
            print('Contatos cadastrados com sucesso!')

        self.export_unsigned_contacts(nao_cadastrados)

    def delete_contacts(self, group_name=None, contact_id_list=None):
        if group_name:
            contacts_cursor = self.client.get_contacts(
                group=group_name
            ).iterfetches(retry_on_rate_exceed=True)

            contacts = []
            for register in contacts_cursor:
                for contact in register:
                    contacts.append(contact.uuid)

            grouped_list = self.break_list_into_groups(contacts)
            for contacts_list in grouped_list:

                try:
                    self.client.bulk_delete_contacts(contacts=contacts_list)
                except Exception as e:
                    pass

        else:
            grouped_list = self.break_list_into_groups(contact_id_list)

            for contacts_list in grouped_list:

                try:
                    self.client.bulk_delete_contacts(contacts=contacts_list)
                except Exception as e:
                    pass



import os
from movva_tools.models.api_token_model import RapidProApiToken

from movva_tools.services.base_service import BaseService
from temba_client.v2 import TembaClient

from movva_tools.exceptions import (
    ObjectDoesNotExistException
)


class APITokenService(BaseService):

    def __init__(self, db_connection=None) -> None:

        super().__init__(db_connection)

        # model table entities
        self.APIToken = RapidProApiToken

    def get_token_by_org_id(self, org_id):

        api_token = self.db_connection.session.query(
            self.APIToken
        ).filter_by(
            is_active=True,
            org_id=org_id
        ).first()

        if api_token:
            return api_token.key
        else:
            raise ObjectDoesNotExistException(
                dababase_object=self.APIToken
            )

    def rapidpro_api_client(self, token):
        try:
            return TembaClient(
                host=os.environ.get('RAPIDPRO_URL'),
                token=token
            )
        except Exception:
            raise Exception(
                'Não foi possível estabelecer conexão com o RapidPro.'
            )



from movva_tools.databases import Base

from sqlalchemy import (
    Boolean, Column, DateTime,
    Integer, String,
)


class RapidProUser(Base):

    __tablename__ = 'auth_user'

    id = Column(Integer, primary_key=True)
    password = Column(String)
    last_login = Column(DateTime)
    is_superuser = Column(Boolean)
    username = Column(String)
    first_name = Column(String)
    last_name = Column(String)
    email = Column(String)
    is_staff = Column(Boolean)
    is_active = Column(Boolean)
    date_joined = Column(DateTime)

    def __repr__(self):
        return f'id:{self.id} username: {self.username} <{self.email}>'



from movva_tools.models.flow_models import RapidProFlowsRevision
from movva_tools.serializers.base_serializer import (
    BaseSerializer, serializer_fields, date_field_format
)


class FlowRevisionSerializer(BaseSerializer):

    class Meta:
        model = RapidProFlowsRevision

    revision = serializer_fields.Integer(dump_only=True)
    created_on = date_field_format
    is_active = serializer_fields.Boolean(dump_only=True)



from movva_tools.models.organization_models import RapidProOrganization
from movva_tools.serializers.base_serializer import (
    BaseSerializer, serializer_fields
)


class OrganizationSerializer(BaseSerializer):
    class Meta:
        model = RapidProOrganization

    uuid = serializer_fields.UUID(dump_only=True)
    name = serializer_fields.String(dump_only=True)
    slug = serializer_fields.String(dump_only=True)



from movva_tools.base_exception import BaseCustomException
from movva_tools.constants import ExceptionMessages


class ObjectDoesNotUpdatedExeption(BaseCustomException):

    def __init__(self, dababase_object) -> None:
        database_object_name = dababase_object.__name__
        self.message = ExceptionMessages.OBJECT_DOES_NOT_UPDATED.format(
            database_object_name
        )
        super().__init__(self.message)


class ObjectDoesNotExistException(BaseCustomException):
    def __init__(self, dababase_object) -> None:
        database_object_name = dababase_object.__name__
        self.message = ExceptionMessages.OBJECT_DOES_NOT_EXISTS.format(
            database_object_name
        )
        super().__init__(self.message)


class FlowNameLongerException(BaseCustomException):

    def __init__(self, max_lenght) -> None:
        self.message = ExceptionMessages.FLOW_NAME_LONGER.format(max_lenght)
        super().__init__(self.message)


class FlowNameStartsOrEndWithWhitespaceException(BaseCustomException):

    def __init__(self) -> None:
        self.message = ExceptionMessages.FLOW_NAME_STARTS_OR_END_WITH_WHITESPACE
        super().__init__(self.message)


class FlowNameInvalidCharacterException(BaseCustomException):

    def __init__(self, invalid_char) -> None:
        self.message = ExceptionMessages.FLOW_INVALID_CHARACTERS.format(invalid_char)
        super().__init__(self.message)


class FlowNameContainsNullCharacterException(BaseCustomException):

    def __init__(self) -> None:
        self.message = ExceptionMessages.FLOW_NAME_CONTAINS_NULL_CHARACTER
        super().__init__(self.message)



from datetime import datetime
import uuid

from movva_tools.databases import Base
from movva_tools.constants import EventMode
from sqlalchemy import (
    UUID, Boolean, Column, DateTime,
    Integer, String,
)


class RapidProCampaigns(Base):
    __tablename__ = 'campaigns_campaign'

    id = Column(Integer, primary_key=True)
    is_active = Column(Boolean)
    created_on = Column(DateTime, default=datetime.utcnow)
    modified_on = Column(DateTime, default=datetime.utcnow)
    uuid = Column(UUID(as_uuid=True), default=uuid.uuid4)
    name = Column(String)
    is_archived = Column(Boolean)
    created_by_id = Column(Integer)
    group_id = Column(Integer)
    modified_by_id = Column(Integer)
    org_id = Column(Integer)

    def __init__(self, created_by_id, org_id, group_id, name):

        super().__init__()
        self.created_by_id = created_by_id
        self.modified_by_id = created_by_id
        self.org_id = org_id
        self.group_id = group_id
        self.name = name
        self.is_archived = False
        self.is_active = True

    def __repr__(self):
        return f'<{self.uuid}:{self.name}>'


class RapidProCampaignEvents(Base):
    __tablename__ = 'campaigns_campaignevent'

    id = Column(Integer, primary_key=True)
    is_active = Column(Boolean)
    created_on = Column(DateTime, default=datetime.utcnow)
    modified_on = Column(DateTime, default=datetime.utcnow)
    uuid = Column(UUID(as_uuid=True), default=uuid.uuid4)
    offset = Column(Integer)
    unit = Column(String)
    start_mode = Column(String)
    event_type = Column(String)
    message = Column(String)
    delivery_hour = Column(Integer)
    campaign_id = Column(Integer)
    created_by_id = Column(Integer)
    flow_id = Column(Integer)
    modified_by_id = Column(Integer)
    relative_to_id = Column(Integer)

    def __init__(
        self, created_by_id,
        offset, unit, event_type,
        campaign_id, flow_id,
        relative_to_id,
        message='', delivery_hour=-1,
    ):
        self.created_by_id = created_by_id
        self.modified_by_id = created_by_id
        self.is_active = True
        self.start_mode = EventMode.INTERRUPT

        self.offset = offset
        self.unit = unit
        self.event_type = event_type

        self.campaign_id = campaign_id
        self.flow_id = flow_id
        self.relative_to_id = relative_to_id
        self.message = message
        self.delivery_hour = delivery_hour

    def __repr__(self):
        return f'<{self.uuid}>'



from movva_tools.services.base_service import BaseService
from movva_tools.models.contacts_models import (
    RapidProContactFields,
    RapidProContactGroups, RapidProContactGroupsContacts
)
from movva_tools.exceptions import ObjectDoesNotExistException


class ContactService(BaseService):

    def __init__(self, db_connection=None) -> None:

        super().__init__(db_connection)

        # model table entities
        self.ContactFields = RapidProContactFields
        self.ContactGroups = RapidProContactGroups
        self.ContactGroupsContacts = RapidProContactGroupsContacts

    def get_contact_by_id(self, id):
        pass

    def get_contact_group_by_id(self, id):
        contact_group = self.db_connection.session.query(
            self.ContactGroups
        ).filter_by(
            id=id
        ).first()

        if contact_group:
            return contact_group
        else:
            raise ObjectDoesNotExistException(
                dababase_object=self.ContactGroups
            )

    def get_contact_group_by_name(self, name, org_id):
        contact_group = self.db_connection.session.query(
            self.ContactGroups
        ).filter_by(
            name=name,
            org_id=org_id
        ).first()

        if contact_group:
            return contact_group
        else:
            raise ObjectDoesNotExistException(
                dababase_object=self.ContactGroups
            )

    def get_contacts_from_groups_by_group_id(self, id):
        pass

    def get_contact_groups_contact_by_contact_id(self, contact_id):
        pass

    def get_contacfield_by_label(self, org, label):
        contact_field = self.db_connection.session.query(
            self.ContactFields
        ).filter_by(
            label=label,
            org_id=org.id
        ).first()

        if contact_field:
            return contact_field
        else:
            raise ObjectDoesNotExistException(
                dababase_object=self.ContactFields
            )



from movva_tools.models.contacts_models import RapidProContactGroups
from movva_tools.serializers.base_serializer import (
    BaseSerializer, serializer_fields
)


class ContactGroupSerializer(BaseSerializer):

    class Meta:
        model = RapidProContactGroups

    uuid = serializer_fields.UUID(dump_only=True)
    name = serializer_fields.String(dump_only=True)
    is_active = serializer_fields.Boolean(dump_only=True)



from abc import ABC, abstractmethod
import os
from urllib.parse import quote_plus

from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from movva_tools.constants import READ_ONLY_ISOLATION_LEVEL_CONNECTION


Base = declarative_base()


class BaseRapidProDatabase(ABC):

    def __init__(self) -> None:
        self._database_url = ''
        self.engine = None
        self.session = None

    def set_connection(self):

        try:
            encoded_password = quote_plus(os.environ.get('DB_PASSWORD'))
            user = os.environ.get('DB_USER')
            host = os.environ.get('DB_HOST')
            port = os.environ.get('DB_PORT')
            db_name = os.environ.get('DB_NAME')
            self._database_url = f"postgresql://{user}:{encoded_password}@{host}:{port}/{db_name}" # noqa
        except Exception:
            raise ConnectionError(
                'As informações de conexão com o banco de dados não foram encontradas.' # noqa
            )

    def set_connection_isolation_level(self, database_url: str, read_only=True) -> str:
        if read_only:
            return database_url+READ_ONLY_ISOLATION_LEVEL_CONNECTION

        return database_url

    def execute_query(self, query, params=None):
        try:
            if not self.engine:
                self.connect()
            conn = self.engine.connect()

            if params is None:
                response = conn.execute(query)
            else:
                response = conn.execute(query, **params)

            conn.close()  # Certifique-se de fechar a conexão
            return response
        except Exception as e:
            conn.close()
            raise e

    def set_session(self):
        if not self.engine:
            self.set_connection()
            self.connect()

        Session = sessionmaker(bind=self.engine)

        self.session = Session()

    @abstractmethod
    def connect(self):
        raise NotImplementedError


class RapidProDatabase(BaseRapidProDatabase):

    def connect(self):
        self.set_connection()
        self._database_url = self.set_connection_isolation_level(
            read_only=False, database_url=self._database_url
        )

        self.engine = create_engine(self._database_url)


class RapidProReadOnlyDatabase(BaseRapidProDatabase):

    def connect(self):
        self.set_connection()
        self.__database_url = self.set_connection_isolation_level(
            read_only=True, database_url=self.__database_url
        )

        self.engine = create_engine(self.__database_url)



from marshmallow import Schema, fields

serializer_fields = fields
date_field_format = serializer_fields.DateTime(
    dump_only=True, format='%d/%m/%Y %H:%M:%S'
)


class BaseSerializer(Schema):
    pass



import uuid
from movva_tools.databases import Base
from datetime import datetime

from sqlalchemy import (
    Boolean, Column, DateTime,
    Integer, String, UUID
)


class RapidProMessages(Base):
    __tablename__ = 'msgs_msg'

    id = Column(Integer, primary_key=True)
    uuid = Column(UUID(as_uuid=True), default=uuid.uuid4)
    text = Column(String)
    high_priority = Column(Boolean, default=False)
    created_on = Column(DateTime, default=datetime.utcnow)
    modified_on = Column(DateTime, default=datetime.utcnow)
    sent_on = Column(DateTime, default=datetime.utcnow)
    queued_on = Column(DateTime, default=datetime.utcnow)
    direction = Column(String)
    status = Column(String)
    visibility = Column(String)
    msg_type = Column(String)
    msg_count = Column(Integer)
    error_count = Column(Integer)
    next_attempt = Column(DateTime, default=datetime.utcnow)
    external_id = Column(UUID(as_uuid=True))
    attachments = Column(String)
    msg_metadata = Column('metadata', String)
    broadcast_id = Column(Integer)
    channel_id = Column(Integer)
    contact_id = Column(Integer)
    contact_urn_id = Column(Integer)
    org_id = Column(Integer)
    topup_id = Column(Integer)
    failed_reason = Column(String)
    flow_id = Column(Integer)

    def __repr__(self):
        return f'id: {self.id} <{self.uuid}>'




from abc import ABC, abstractmethod
import os
import requests


class BaseSlackNotification(ABC):
    """
        Classe base que implementa os métodos fundamentais e configuração de webhook
        para enviar as notificações ao canal do slack.
    """

    def __init__(
        self, title, data,
        emoji=True,
        extra_params=None
    ):
        self.title = title
        self.emoji = emoji
        self.data = data
        self.extra_params = extra_params
        self.response = None

        self._payload = {}
        self._webhook_url = os.environ.get('SLACK_WEBHOOK_URL')
        self._app = 'Cópia e Agendamento de Fluxos'

    @property
    def payload(self):
        """
            Retorna o payload que será enviado na requisição ao webhook do slack.
        """
        return self._payload

    def slack_payload_initialization(self):
        self._payload.update(
            {'blocks': []}
        )

    def slack_payload_header(self):

        header_obj = {
            "type": "header",
            "text": {
                "type": "plain_text",
                "text": self.title,
                "emoji": self.emoji
            }
        }

        return header_obj

    def slack_application_content(self):
        return {
            "type": "section",
            "fields": [
                {
                    "type": "mrkdwn",
                    "text": f"*Aplicação:* {self._app}"
                }
            ]
        }

    def validate_content_fields(self):

        mandatory_keys = ['type']
        declared_fields = ['fields', 'text']

        error = ''
        for payload_content in self._payload['blocks']:

            if not set(mandatory_keys).issubset(set(payload_content.keys())) or type(payload_content) != dict: # noqa
                error += 'Content must be a dict and must have "type" as key\n'
            if not any(field in declared_fields for field in payload_content):
                error += f'No declared fields found. Choices are: {declared_fields}'  # noqa
            if error:
                raise Exception(error)

    def send_slack_message(self):
        headers = {
            "content-type": "application/json"
        }
        data = dict(
            text='' if not self.extra_params else self.extra_params,
            **self._payload
        )

        response = requests.post(
            self._webhook_url,
            json=data,
            headers=headers,
            timeout=25
        )
        return response

    @abstractmethod
    def _build_notification_content(self):
        """
            Método que implementa as particularidades dos alertas de
            acordo com os parâmetros passados.
            - Deve ser obrigatoriamente sobrescrito.
        """
        raise NotImplementedError('This method must be overwritten.')

    def send_notification(self):

        self.slack_payload_initialization()

        self._build_notification_content()

        self.validate_content_fields()

        self.response = self.send_slack_message()


class SlackFlowSchedulledNotification(BaseSlackNotification):

    def slack_organization_content(self):
        return {
            "type": "section",
            "fields": [
                {
                    "type": "mrkdwn",
                    "text": f"*Organização:* {self.data['organization']['name']}\n *slug:* {self.data['organization']['slug']}" # noqa
                }
            ]
        }

    def slack_flow_content(self):
        return {
            "type": "section",
            "fields": [
                {
                    "type": "mrkdwn",
                    "text": f"*Fluxo:* {self.data['flow']['name']}\n *criado em:* {self.data['flow']['created_on']}" # noqa
                }
            ]
        }

    def slack_flow_revision_content(self):
        return {
            "type": "section",
            "fields": [
                {
                    "type": "mrkdwn",
                    "text": f"*Revisão do Fluxo:* {self.data['flow_revision']['revision']}\n *criado em:* {self.data['flow_revision']['created_on']}" # noqa
                }
            ]
        }

    def slack_campaign_content(self):
        return {
            "type": "section",
            "fields": [
                {
                    "type": "mrkdwn",
                    "text": f"*Campanha:* {self.data['campaign']['name']}\n *criada em:* {self.data['flow']['created_on']}" # noqa
                }
            ]
        }

    def slack_campaign_event_content(self):
        return {
            "type": "section",
            "fields": [
                {
                    "type": "mrkdwn",
                    "text": f"*Agendamento da Campanha:* {self.data['campaign_event']['config_time']}\n *criado em:* {self.data['campaign_event']['created_on']}" # noqa
                }
            ]
        }

    def slack_group_content(self):
        return {
            "type": "section",
            "fields": [
                {
                    "type": "mrkdwn",
                    "text": f"*Grupo de Contatos:* {self.data['group']['name']}"  # noqa
                }
            ]
        }

    def _build_notification_content(self):
        self._payload['blocks'].extend(
            [
                self.slack_payload_header(),
                self.slack_application_content(),
                self.slack_organization_content(),
                self.slack_flow_content(),
                self.slack_flow_revision_content(),
                self.slack_campaign_content(),
                self.slack_campaign_event_content(),
                self.slack_group_content()
            ]
        )



import uuid
from datetime import datetime

from sqlalchemy import (
    JSON, UUID, Boolean, Column, DateTime,
    Integer, String, ForeignKey
)

from movva_tools.databases import Base
from sqlalchemy.orm import relationship


class RapidProChannel(Base):
    __tablename__ = 'channels_channel'

    id = Column(Integer, primary_key=True)
    is_active = Column(Boolean, default=True)
    created_on = Column(DateTime, default=datetime.utcnow)
    modified_on = Column(DateTime, default=datetime.utcnow)
    uuid = Column(UUID(as_uuid=True), default=uuid.uuid4)
    channel_type = Column(String)
    name = Column(String)
    address = Column(String)
    country = Column(String)
    claim_code = Column(String)
    secret = Column(String)
    last_seen = Column(DateTime, default=datetime.utcnow)
    device = Column(String, default="")
    os = Column(String, default="")
    alert_email = Column(String, default="")
    config = Column(JSON)
    schemes = Column(String)
    role = Column(String)
    bod = Column(String, default="")
    tps = Column(Integer)
    created_by_id = Column(Integer)
    modified_by_id = Column(Integer)
    org_id = Column(Integer)
    parent_id = Column(Integer, default="")

    def __repr__(self):
        return f'uuid: {self.uuid} <{self.name}|{self.address}>'


class RapidProChannelLog(Base):
    __tablename__ = 'channels_channellog'

    id = Column(Integer, primary_key=True)
    description = Column(String)
    url = Column(String)
    method = Column(String)
    request = Column(String)
    response = Column(String)
    response_status = Column(Integer)
    created_on = Column(DateTime, default=datetime.utcnow)
    request_time = Column(Integer)
    channel_id = Column(Integer, ForeignKey('channels_channel.id'))
    connection_id = Column(Integer)
    msg_id = Column(Integer)

    # Defina o relacionamento com RapidProChannel
    channel = relationship('RapidProChannel')

    def __repr__(self):
        return f'id: {self.id} <channel: {self.channel.name}>'



import json
import os
import re
import pandas as pd

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from movva_tools.constants import BaseEnum, GoogleSheets
from movva_tools.utils import (
    replace_newline_character_for_space, strip_whitespaces
)


# If modifying scopes, delete the file token.json.


class GoogleSheetsIntegration:

    def __init__(self, link, spreadsheet_page):
        self._link = link
        self.spreadsheet_page = spreadsheet_page
        self.service = None
        self._secret_manager_service = None
        self._spreadsheet_id = None
        self._spreadsheet_data = None
        self._scope = GoogleSheets.READ_ONLY_SCOPE

    def check_data(self):
        """
            Verifica se as informações necessárias para a classe
            foram informadas.
        """
        if not self._link or not self._spreadsheet_id or not self.spreadsheet_page:
            raise Exception('The spreadsheet link, name and id must be informed')

    def __extract_id_from_google_sheets_link(self):
        """
            Extrai o id do link da planilha do Google através do
            padrão de expressão regular
        """

        pattern = r'/d/([a-zA-Z0-9-_]+)'

        # Procura pelo padrão no link fornecido
        match = re.search(pattern, self._link)

        # Se houver uma correspondência, retorna o ID encontrado, caso contrário, retorna None
        if match:
            return match.group(1)
        else:
            raise Exception('Planilha não encontrada.')

    def __fetch_credentials_data(self):

        environment = None
        token_path = None
        creds_path = os.environ.get('CREDENTIALS_API_GOOGLE_SHEETS', None)

        if not creds_path:
            token_path = os.environ.get('TOKEN_LOCAL_PATH', '')
            creds_path = os.environ.get('CREDENTIALS_LOCAL_PATH', None)
            environment = 'local'
        else:
            token_path = os.environ.get('TOKEN_API_GOOGLE_SHEETS', None)
            environment = 'production'

        if not creds_path or not token_path:
            raise ConnectionError('As credenciais da API do Google Sheets não foram informadas.')

        return creds_path, token_path, environment

    def _validate_google_sheets_token(self, creds, creds_path, gsheets_token_path, environment):
        if environment == 'local':
            # The file token.json stores the user's access and refresh tokens, and is
            # created automatically when the authorization flow completes for the first
            # time.

            # If there are no (valid) credentials available, let the user log in.
            if not creds or not creds.valid:
                # if creds and creds.expired and creds.refresh_token:
                #     creds.refresh(Request())
                # else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    creds_path, self._scope
                )
                creds = flow.run_local_server(port=0)
                # Save the credentials for the next run
                print('Openning token.json file...')
                with open(gsheets_token_path, 'w') as token:
                    token.write(creds.to_json())

        else:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())

        return creds

    def __authorize_google_sheets(self):
        """
            Cria/carrega credenciais de acesso da API do Google Sheets
            e inicia a conexão com a API.
        """
        creds_path, token_path, environment = self.__fetch_credentials_data()
        creds = None

        if environment != 'local':

            gsheets_token_path = json.loads(
                token_path
            )

            creds = Credentials.from_authorized_user_info(
                gsheets_token_path, self._scope
            )
        else:
            gsheets_token_path = token_path

            if os.path.exists(gsheets_token_path):
                creds = Credentials.from_authorized_user_file(
                    gsheets_token_path, self._scope
                )

        creds = self._validate_google_sheets_token(
            creds, creds_path, gsheets_token_path, environment
        )

        # Create Google Sheets connection
        service = build('sheets', 'v4', credentials=creds)

        return service

    def read_sheet_data(self, end_selection: str = None) -> list:
        """
            Lê e armazena em uma estrutura de listas o conteúdo da planilha.
        """

        self._spreadsheet_id = self.__extract_id_from_google_sheets_link()

        self.check_data()

        self.service = self.__authorize_google_sheets()

        # Call the Sheets API
        end_selection = end_selection if end_selection else 'BZ'
        sheet = self.service.spreadsheets()
        result = sheet.values().get(
            spreadsheetId=self._spreadsheet_id,
            range=f'{self.spreadsheet_page}!A1:{end_selection}'
        ).execute()

        values = result.get('values', [])

        if not values:
            raise Exception('Planilha sem informação.')

        self._spreadsheet_data = values

        return values


class FlowGoogleSheets(GoogleSheetsIntegration):

    def __init__(self, link, spreadsheet_page):
        super().__init__(link, spreadsheet_page)
        self.message_columns = None
        self._scope = GoogleSheets.DEFAULT_SCOPE

    class FlowColumns(BaseEnum):
        ORGANIZATION = 'Organização'
        ORGANIZATION_OF_TEMPLATE = 'Organização do template a copiar'
        CAMPAING_NAME = 'Campanha'
        FLOW = 'Nome do fluxo'
        SEND_DATE = 'Data de Envio'
        SEND_DAY = 'Dia da semana'
        SEND_TIME = 'Horário de Envio (Brasil)'
        REFERENCE_DATE = 'Data de referência'
        REFERENCE_HOUR = 'Hora de referência'
        DAYS_FROM_REFERENCE_DATE = 'Dias da data de referência'
        HOURS_FROM_REFERENCE_DATE = 'Horas da data de referência'
        LABEL_REFERENCE_DATE = 'Label Data de referência'
        MESSAGE_TYPE = 'Tipo de mensagem'
        TEMPLATE_FLOW = 'Template a copiar'
        MESSAGE_PATTERN = r'Mensagem#\d+'
        CONTACT_GROUP = 'Grupo'
        UPLOAD = 'Subir?'

        CONTROL_TO_UPLOAD = 'FALSO'
        CONTROL_NOT_UPLOADED = 'VERDADEIRO'

    def parse_no_newline_columns(self, columns: list):
        return [
            replace_newline_character_for_space(column) for column in columns
        ]

    def validate_message_columns(self, columns: list):
        message_columns_pattern = self.FlowColumns.MESSAGE_PATTERN

        flow_messages_list = [
            item for item in columns
            if re.search(message_columns_pattern, str(item))
        ]

        if not flow_messages_list:
            raise Exception('Não foram informadas mensagens.')

        self.message_columns = flow_messages_list
        return flow_messages_list

    def validate_spreadsheet_columns(self, columns: list):

        columns_messages = self.validate_message_columns(columns)

        flowcolumns_list = set(columns).difference(columns_messages)

        validation_columns = [
            column.value for column in self.FlowColumns
            if column.value not in
            (
                self.FlowColumns.MESSAGE_PATTERN,
                self.FlowColumns.CONTROL_TO_UPLOAD,
                self.FlowColumns.CONTROL_NOT_UPLOADED
            )
        ]

        try:
            assert all(
                column in flowcolumns_list for column in validation_columns
            )
        except Exception:
            raise Exception('As colunas esperadas da planilha não foram encontradas.')

    def read_sheet_data(self, end_selection: str = None) -> list:
        values = super().read_sheet_data(end_selection)
        columns = [strip_whitespaces(column) for column in values[0]]
        columns = self.parse_no_newline_columns(columns)
        data = pd.DataFrame(values[1:], columns=columns)

        self._spreadsheet_data = data
        self.validate_spreadsheet_columns(data.columns)

        return data

    def sheets_column(self, colunas_dataframe: int):
        if colunas_dataframe <= 0:
            raise ValueError("O número de colunas deve ser maior que zero.")

        coluna_sheets = ""
        while colunas_dataframe > 0:
            colunas_dataframe, remainder = divmod(colunas_dataframe - 1, 26)
            coluna_sheets = chr(remainder + ord('A')) + coluna_sheets

        return coluna_sheets

    def write_sheet_data(self, rows: list, value: str, control_column: str):
        """
            Escreve um valor em uma coluna específica para linhas específicas
            da planilha.
        """
        self.check_data()

        if not self.service:
            self.service = self.__authorize_google_sheets()

        sheet = self.service.spreadsheets()

        template_rows = [item['template'] for item in rows]
        flow_rows = [item['flow'] for item in rows]

        selection = template_rows + flow_rows
        selection.sort()

        # Escrever nas linhas específicas da coluna 'Feito'
        self._spreadsheet_data.loc[
            self._spreadsheet_data.index.isin(selection), control_column
        ] = value

        columns = len(self._spreadsheet_data.columns)
        cell = self.sheets_column(colunas_dataframe=columns)

        control_values = list(self._spreadsheet_data[self.FlowColumns.UPLOAD].values)
        values = []
        for value in control_values:
            values.append(
                [value]
            )

        # Escrever o DataFrame de volta na planilha
        sheet.values().update(
            spreadsheetId=self._spreadsheet_id,
            range=f'{self.spreadsheet_page}!{cell}2',
            body={'values': values},
            valueInputOption='RAW'
        ).execute()

        print(
            f'Valor "{value}" escrito na coluna "{control_column}".'
        )



from movva_tools.serializers.campaign_event_serializer import CampaingEventSerializer
from movva_tools.serializers.campaign_serializer import CampaignSerializer
from movva_tools.serializers.organization_serializer import OrganizationSerializer
from movva_tools.serializers.flow_serializer import FlowSerializer
from movva_tools.serializers.contact_group_serializer import ContactGroupSerializer
from movva_tools.serializers.flow_revision_serializer import FlowRevisionSerializer



from datetime import timedelta, datetime
from movva_tools.constants import EventUnitTime, EventMode, EventType

from movva_tools.models.campaign_models import (
    RapidProCampaignEvents, RapidProCampaigns
)
from movva_tools.models.flow_models import RapidProFlows
from movva_tools.models.user_models import RapidProUser
from movva_tools.models.organization_models import RapidProOrganization
from movva_tools.models.contacts_models import (
    RapidProContactFields, RapidProContactGroups
)

from movva_tools.services.contacts_service import ContactService
from movva_tools.services.base_service import BaseService


class CampaingService(BaseService):

    def __init__(self, db_connection=None) -> None:

        super().__init__(db_connection)

        # services
        self.contact_service = ContactService(
            db_connection=self.db_connection
        )

        # model table entities
        self.Campaign = RapidProCampaigns
        self.CampaignEvent = RapidProCampaignEvents

    def create_campaign(
        self,
        org: RapidProOrganization,
        user: RapidProUser,
        name: str,
        group: RapidProContactGroups
    ):

        campaign = self.Campaign(
            org_id=org.id,
            created_by_id=user.id,
            name=name,
            group_id=group.id
        )

        self.add(campaign)
        self.flush()

        return campaign

    def define_delivery_hour_campaign_event(
        self, offset, reference_date, delivery_time
    ):
        return reference_date + timedelta(
            days=offset, hours=delivery_time.hour, minutes=delivery_time.minute
        )

    def fetch_reference_date_field(self, label, organization):

        return self.contact_service.get_contacfield_by_label(
            org=organization,
            label=label
        )

    def set_config_campaign_event(
        self,
        offset, offset_hour,
        send_time, reference_date,
        reference_hour,
    ):
        delivery_time = None
        unit = EventUnitTime.DAYS

        """
            Hora de envio com minutos - deslocamento em minutos
            Não é informada hora de envio - deslocamento em dias
            Informada hora de envio sem minutos - deslocamento em dia e hora
        """
        if send_time and send_time >= reference_hour:

            if send_time.minute != 0:
                unit = EventUnitTime.MINUTES
                reference_date_with_time = reference_date + timedelta(hours=reference_hour.hour, minutes=reference_hour.minute)
                offset = (offset * 24 * 60) + (offset_hour.hour * 60) + send_time.minute
                send_time = -1
                delivery_time = reference_date_with_time + timedelta(minutes=offset)
            else:
                send_time = send_time.hour
                delivery_time = reference_date + timedelta(days=offset)

        elif send_time and send_time < reference_hour:
            if send_time.minute != 0:
                unit = EventUnitTime.MINUTES
                reference_date_with_time = reference_date + timedelta(hours=reference_hour.hour, minutes=reference_hour.minute)
                offset = (offset * 24 * 60) - (offset_hour.hour * 60) - send_time.minute
                send_time = -1
                delivery_time = reference_date_with_time + timedelta(minutes=offset)
            else:
                send_time = send_time.hour
                delivery_time = reference_date + timedelta(days=offset) - timedelta(hours=offset_hour.hour)

        elif not send_time:
            send_time = -1
            delivery_time = reference_date + timedelta(days=offset)

        elif not delivery_time:
            raise Exception('Não foi possível configurar horário de envio.')

        return {
            'delivery_hour': send_time,
            'unit': unit,
            'event_type': EventType.FLOW,
            'offset': offset,
            'delivery_time': delivery_time
        }

    def create_campaign_event(
        self,
        campaign: RapidProCampaigns,
        flow: RapidProFlows,
        relative_to: RapidProContactFields,
        configuration: dict
    ):

        campaign_event = self.CampaignEvent(
            campaign_id=campaign.id,
            created_by_id=campaign.created_by_id,
            flow_id=flow.id,
            relative_to_id=relative_to.id,
            delivery_hour=configuration['delivery_hour'],
            unit=configuration['unit'],
            event_type=configuration['event_type'],
            offset=configuration['offset']
        )

        self.add(campaign_event)
        self.flush()

        return campaign_event



from movva_tools.databases import Base

from sqlalchemy import (
    UUID, Boolean, Column, DateTime,
    Integer
)


class RapidProApiToken(Base):
    __tablename__ = 'api_apitoken'

    is_active = Column(Boolean, default=True)
    key = Column(UUID(as_uuid=True), primary_key=True)
    created = Column(DateTime)
    org_id = Column(Integer)
    role_id = Column(Integer)
    user_id = Column(Integer)

    def __repr__(self):
        return f'Api Token: {self.key} <org:{self.org.id}>'



from functools import wraps
from psycopg2 import ProgrammingError


def ensure_transaction(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        db_session_instance = self.db_connection.session
        try:
            result = func(self, *args, **kwargs)
            db_session_instance.commit()  # Commit a transação se tudo estiver correto
            return result
        except (Exception, TypeError, ProgrammingError, AttributeError) as e:
            db_session_instance.rollback()  # Reverta a transação em caso de exceção
            raise e  # Relevante a exceção após o rollback
    return wrapper



from movva_tools.databases import RapidProDatabase
from sqlalchemy.exc import SQLAlchemyError


class BaseService:

    def __init__(self, db_connection=None) -> None:
        self.db_connection = db_connection or self.__create_default_connection()

    def __create_default_connection(self):
        db = RapidProDatabase()
        db.set_session()
        return db

    def add(self, object):
        self.db_connection.session.add(object)

    def commit(self, object):
        self.db_connection.session.commit()

    def flush(self):
        """
            Flush para obter o ID gerado sem commit
        """
        self.db_connection.session.flush()

    def rollback_session(self):
        try:
            self.db_connection.session.rollback()
            return True
        except SQLAlchemyError as e:
            raise Exception(e)

    def close_session(self):
        self.db_connection.session.close()



import json
import pandas as pd
from sqlalchemy import desc
from typing import Tuple

from movva_tools.models.campaign_models import (
    RapidProCampaignEvents, RapidProCampaigns
)
from movva_tools.models.channel_models import RapidProChannel
from movva_tools.notifications import SlackFlowSchedulledNotification
from movva_tools.models.user_models import RapidProUser
from movva_tools.models.organization_models import RapidProOrganization
from movva_tools.models.contacts_models import RapidProContactGroups
from movva_tools.models.flow_models import (
    RapidProFlowChannelDependencies, RapidProFlowFieldDependencies,
    RapidProFlowFlowLabelDependencies, RapidProFlowGroupDependencies,
    RapidProFlowLabel, RapidProFlows,
    RapidProFlowsFlowLabels, RapidProFlowsRevision
)

from movva_tools.serializers import CampaingEventSerializer
from movva_tools.serializers import ContactGroupSerializer
from movva_tools.serializers import FlowRevisionSerializer
from movva_tools.serializers import OrganizationSerializer
from movva_tools.serializers import CampaignSerializer
from movva_tools.serializers import FlowSerializer

from movva_tools.services.base_service import BaseService
from movva_tools.services.api_token_service import APITokenService
from movva_tools.services.campaign_service import CampaingService
from movva_tools.services.contacts_service import ContactService
from movva_tools.services.organization_service import OrganizationService
from movva_tools.services.user_service import UserService

from movva_tools.constants import MAX_NAME_LEN
from movva_tools.constants import Flow, FlowType, GoFlowTypes
from movva_tools.decorators import ensure_transaction
from movva_tools.exceptions import ObjectDoesNotExistException

from movva_tools.utils import DATE_FORMAT, TIME_FORMAT, json_representation, serialize
from movva_tools.utils import break_list_into_groups, datetime_string_parser
from movva_tools.validators import FlowNameValidator
from movva_tools.integrations.google_integration import FlowGoogleSheets


class FlowService(BaseService):

    def __init__(self, db_connection=None) -> None:

        super().__init__(db_connection=db_connection)

        # model table entities
        self.flow = RapidProFlows
        self.flow_revision = RapidProFlowsRevision
        self.org = RapidProOrganization

        # services
        self.user_service = UserService(
            db_connection=self.db_connection
        )
        self.organization_service = OrganizationService(
            db_connection=self.db_connection
        )

    def __set_default_flow_metadata(self):

        return {
            "results": [],
            "dependencies": [],
            "waiting_exit_uuids": [],
            "parent_refs": []
        }

    def __set_initial_flow_revision_definition(self, flow: RapidProFlows):
        return {
            Flow.DEFINITION_NAME: flow.name,
            Flow.DEFINITION_UUID: str(flow.uuid),
            Flow.DEFINITION_SPEC_VERSION: Flow.CURRENT_SPEC_VERSION,
            Flow.DEFINITION_LANGUAGE: Flow.BASE_LANGUAGE,
            Flow.DEFINITION_TYPE: GoFlowTypes.TYPE_MESSAGE,
            Flow.DEFINITION_NODES: [],
            Flow.DEFINITION_UI: {},
            Flow.DEFINITION_REVISION: 1,
            Flow.DEFINITION_EXPIRE_AFTER_MINUTES: Flow.EXPIRES_AFTER_MINUTES
        }

    def __set_flow_revision_definition(
        self, flow: RapidProFlows, setted_definition: json
    ):

        setted_definition[Flow.DEFINITION_NAME] = flow.name
        setted_definition[Flow.DEFINITION_REVISION] = 1
        setted_definition[Flow.DEFINITION_UUID] = str(flow.uuid)

        return setted_definition

    def is_valid_name(cls, name: str) -> bool:
        try:
            FlowNameValidator()(value=name)
            return True
        except Exception:
            return False

    def get_flow_by_name(self, flow_name: str):

        flow = self.db_connection.session.query(self.flow).filter_by(
            name=flow_name
        ).first()

        if flow:
            return flow
        else:
            raise ObjectDoesNotExistException(dababase_object=self.flow)

    def get_flow_by_id(self, flow_id: int):

        flow = self.db_connection.session.query(self.flow).filter_by(
            id=flow_id
        ).first()

        if flow:
            return flow
        else:
            raise ObjectDoesNotExistException(dababase_object=self.flow)

    def create_flow_revision(
        self, flow: RapidProFlows, user: RapidProUser,
        setted_definition: json = None
    ):
        if setted_definition:
            definition = self.__set_flow_revision_definition(
                flow, setted_definition=setted_definition
            )
        else:
            definition = self.__set_initial_flow_revision_definition(flow=flow)

        new_flow_revision = self.flow_revision(
            created_by_id=user.id,
            flow_id=flow.id,
            definition=definition
        )

        self.add(new_flow_revision)
        self.flush()
        return new_flow_revision

    def _create_flow(
        self,
        user: RapidProUser,
        org: RapidProOrganization,
        flow_name: str,
        flow_type=FlowType.TYPE_MESSAGE,
        expires_after_minutes=Flow.EXPIRES_AFTER_MINUTES
    ):

        new_flow = self.flow(
            org_id=org.id,
            created_by_id=user.id,
            name=flow_name,
            flow_type=flow_type,
            expires_after_minutes=expires_after_minutes,
            flow_metadata=self.__set_default_flow_metadata()
        )

        self.add(new_flow)
        self.flush()
        return new_flow

    def create_flow(
        self,
        org: RapidProOrganization,
        user: RapidProUser,
        flow_name: str,
        external_context: bool = True
    ):

        self.is_valid_name(flow_name)

        return self._create_flow(
            org=org,
            user=user,
            flow_name=flow_name
        )


class FlowLabelService(BaseService):
    def __init__(self, db_connection=None) -> None:

        super().__init__(db_connection)

        # model table entities
        self.Flow = RapidProFlows
        self.Organization = RapidProOrganization
        self.FlowLabel = RapidProFlowLabel
        self.FlowsFlowLabels = RapidProFlowsFlowLabels

        # services
        self.user_service = UserService(
            db_connection=self.db_connection
        )
        self.organization_service = OrganizationService(
            db_connection=self.db_connection
        )
        self.flow_service = FlowService(
            db_connection=self.db_connection
        )

    def get_label_by_id(self, id: int):
        label = self.db_connection.session.query(self.FlowLabel).filter_by(
            id=id
        ).first()

        if label:
            return label
        else:
            raise ObjectDoesNotExistException(dababase_object=self.FlowLabel)

    def get_label_by_name(self, name: str, org_id: int):
        label = self.db_connection.session.query(self.FlowLabel).filter_by(
            name=name,
            org_id=org_id
        ).first()

        if label:
            return label
        else:
            raise ObjectDoesNotExistException(dababase_object=self.FlowLabel)

    def _create_flow_label(
        self, name: str, org: RapidProOrganization, flow: RapidProFlows
    ):

        flow_label = self.FlowLabel(
            org_id=org.id,
            name=name
        )

        self.add(flow_label)
        self.flush()

        flows_flowlabels = self.FlowsFlowLabels(
            flow_id=flow.id,
            flowlabel_id=flow_label.id
        )

        self.add(flows_flowlabels)
        self.flush()

        return flow_label

    def create(self, name: str, org: RapidProOrganization, flow: RapidProFlows):
        return self._create_flow_label(name, org, flow)


class FlowDependenciesService(BaseService):

    def __init__(self, db_connection=None) -> None:

        super().__init__(db_connection)

        # model table entities
        self.FlowChannelDependencies = RapidProFlowChannelDependencies

        self.FlowGroupDependencies = RapidProFlowGroupDependencies

        self.FlowFieldDependencies = RapidProFlowFieldDependencies

        self.FlowFlowLabelDependencies = RapidProFlowFlowLabelDependencies

    def _set_flow_dependencies(
        self,
        flow: RapidProFlows, channel: RapidProChannel,
        group: RapidProContactGroups, flowlabel: RapidProFlowLabel
    ):
        flow_channel_dependencies = self.FlowChannelDependencies(
            flow_id=flow.id,
            channel_id=channel.id
        )
        self.add(flow_channel_dependencies)

        flow_group_dependencies = self.FlowGroupDependencies(
            flow_id=flow.id,
            group_id=group.id
        )
        self.add(flow_group_dependencies)

        flow_flowlabel_dependencies = self.FlowFlowLabelDependencies(
            flow_id=flow.id,
            flowlabel_id=flowlabel.id
        )
        self.add(flow_flowlabel_dependencies)

        self.flush()

    def create(
        self, flow: RapidProFlows, channel: RapidProChannel,
        group: RapidProContactGroups, flowlabel: RapidProFlowLabel
    ):
        return self._set_flow_dependencies(flow, channel, group, flowlabel)


class CopyFlowAndScheduleService(BaseService):

    def __init__(self, db_connection=None) -> None:

        super().__init__(db_connection)

        # spreadsheet reader to copy flow
        self.spreadsheet_reader = FlowGoogleSheets

        # model table entities
        self.flow_revision = RapidProFlowsRevision

        # services
        self.flow_service = FlowService(
            db_connection=self.db_connection
        )
        self.flowlabel_service = FlowLabelService(
            db_connection=self.db_connection
        )
        self.organization_service = OrganizationService(
            db_connection=self.db_connection
        )
        self.user_service = UserService(
            db_connection=self.db_connection
        )
        self.api_token_service = APITokenService(
            db_connection=self.db_connection
        )
        self.contact_service = ContactService(
            db_connection=self.db_connection
        )
        self.campaign_service = CampaingService(
            db_connection=self.db_connection
        )

        # serializers
        self.FlowSerializer = FlowSerializer
        self.CampaignSerializer = CampaignSerializer
        self.OrganizationSerializer = OrganizationSerializer
        self.FlowRevisionSerializer = FlowRevisionSerializer
        self.CampaignEventSerializer = CampaingEventSerializer
        self.ContactGroupSerializer = ContactGroupSerializer

    def get_unique_name(self, org: RapidProOrganization, base_name: str):

        name = f'{base_name[:MAX_NAME_LEN]}' if len(base_name) > MAX_NAME_LEN else base_name
        qs = self.db_connection.session.query(self.flow).filter_by(
            org_id=org.id,
            is_active=True,
            name=name
        ).all()

        if qs:
            name = 'Copy_of ' + name

        return name

    def get_last_flow_revision(self, flow: RapidProFlows):

        flow_revision = self.db_connection.session.query(
            self.flow_revision
        ).order_by(
            desc(self.flow_revision.created_on)
        ).filter_by(
            flow_id=flow.id
        ).first()

        if flow_revision:
            return flow_revision
        else:
            raise ObjectDoesNotExistException(
                dababase_object=self.flow_revision
            )

    def _transfer_flow_revision_messages(
        self,
        json_string: str,
        template_data: dict,
        new_flow_data: dict,
        reader: FlowGoogleSheets
    ):
        # identificar padrão Mensagem#N, onde N é o numero da msg

        messages = zip(
            [template_data[key] for key in reader.message_columns],
            [new_flow_data[key] for key in reader.message_columns]
        )

        for actual, new in messages:
            actual = actual.replace('\n', '\\n')
            new = new.replace('\n', '\\n')
            actual = f'"{actual}"'
            new = f'"{new}"'
            json_string = json_string.replace(actual, new)

        return json_string

    def transfer_flow_revision_data(
        self, flow_json: dict,
        template_data: dict,
        new_flow_data: dict,
        spreadsheet_reader: FlowGoogleSheets
    ):

        # aqui serão chamados os métodos para cada etapa do definition
        flow_data = json.dumps(flow_json, ensure_ascii=False)

        flow_data = self._transfer_flow_revision_messages(
            json_string=flow_data,
            template_data=template_data,
            new_flow_data=new_flow_data,
            reader=spreadsheet_reader
        )
        return json.loads(flow_data)

    def set_rapidpro_client_to_organization(self, org_destination):
        api_token = self.api_token_service.get_token_by_org_id(
            org_id=org_destination.id
        )

        return self.api_token_service.rapidpro_api_client(
            token=api_token
        )

    def import_definition(
        self,
        user: RapidProUser,
        definition: json,
        flow_destination: RapidProFlows,
        spreadsheet_reader: FlowGoogleSheets,
        template_data: dict,
        new_flow_data: dict
    ):
        """
            Allows setting the definition for a flow from another definition.
            All UUID's will be remapped.
        """

        flow_revision_copy = definition.copy()

        # aqui deve entrar as modificações do flow revision
        flow_revision_copy = self.transfer_flow_revision_data(
            template_data=template_data,
            new_flow_data=new_flow_data,
            spreadsheet_reader=spreadsheet_reader,
            flow_json=flow_revision_copy
        )

        flow_revision_destination = self.flow_service.create_flow_revision(
            flow=flow_destination,
            user=user,
            setted_definition=flow_revision_copy
        )

        return flow_revision_destination

    def get_revision_definition(self, flow) -> dict:
        revision = self.get_last_flow_revision(
            flow=flow
        )

        definition = revision.definition

        return json.loads(definition)

    def _check_organization_data(
        self,
        org_flow: RapidProOrganization,
        origin_org: str
    ) -> bool:
        org_flow_name = org_flow.name
        if org_flow_name == origin_org:
            return True

        return False

    def clone_flow(
        self,
        raw_flow: dict, template: dict,
        user: RapidProUser,
        reader: FlowGoogleSheets
    ) -> Tuple[RapidProFlows, RapidProFlowsRevision, RapidProOrganization]:

        origin_organization_name = raw_flow[
            reader.FlowColumns.ORGANIZATION_OF_TEMPLATE
        ]
        destiny_organization_name = raw_flow[
            reader.FlowColumns.ORGANIZATION
        ]
        base_flow_name = raw_flow[reader.FlowColumns.TEMPLATE_FLOW]
        flow_suggested_name = raw_flow[reader.FlowColumns.FLOW]

        org_destination = self.organization_service.get_org_by_name(
            org_name=destiny_organization_name
        )

        base_flow = self.flow_service.get_flow_by_name(
            flow_name=base_flow_name
        )
        org_related_to_flow = self.organization_service.get_org_by_id(
            org_id=base_flow.org_id
        )

        if not self._check_organization_data(
            org_flow=org_related_to_flow,
            origin_org=origin_organization_name
        ):
            raise Exception(
                'A organização de origem não contém o fluxo original informado para cópia.'  # noqa
            )

        copy = self.flow_service.create_flow(
            org=org_destination,
            user=user,
            flow_name=flow_suggested_name
        )

        base_flow_json = self.get_revision_definition(
            flow=base_flow
        )

        revision = self.import_definition(
            user=user,
            definition=base_flow_json,
            flow_destination=copy,
            new_flow_data=raw_flow,
            template_data=template,
            spreadsheet_reader=reader
        )

        return copy, revision, org_destination

    def prepare_flow_copy_spreadsheet(
        self,
        registers: pd.DataFrame,
        reader: FlowGoogleSheets
    ):

        # filter lines that have data not registred.
        registers = registers[
            registers[reader.FlowColumns.UPLOAD] == 'VERDADEIRO'
        ]

        if registers.empty:
            raise Exception('Planilha sem informações.')

        registers = registers.reset_index().to_dict(orient='records')

        return registers

    def create_campaign(
        self,
        reader: FlowGoogleSheets,
        raw_flow: dict,
        flow: RapidProFlows,
        org: RapidProOrganization,
        user: RapidProUser,
        group: RapidProContactGroups
    ) -> Tuple[RapidProCampaigns, RapidProCampaignEvents, dict]:

        spreadsheet_reference_date = None
        spreadsheet_reference_hour = None
        spreadsheet_offset_hour = None
        spreadsheet_offset = None
        spreadsheet_send_time = None
        spreadsheet_label_reference_date = None

        spreadsheet_campaign_name = raw_flow[reader.FlowColumns.CAMPAING_NAME]

        try:
            spreadsheet_label_reference_date = raw_flow[
                reader.FlowColumns.LABEL_REFERENCE_DATE
            ]

            spreadsheet_reference_date = datetime_string_parser(
                date_str=raw_flow[reader.FlowColumns.REFERENCE_DATE],
                pattern=DATE_FORMAT
            )

            spreadsheet_reference_hour = datetime_string_parser(
                date_str=raw_flow[reader.FlowColumns.REFERENCE_HOUR],
                pattern=TIME_FORMAT
            )

            spreadsheet_offset_hour = datetime_string_parser(
                date_str=raw_flow[reader.FlowColumns.HOURS_FROM_REFERENCE_DATE],
                pattern=TIME_FORMAT
            )

            spreadsheet_offset = int(
                raw_flow[reader.FlowColumns.DAYS_FROM_REFERENCE_DATE]
            )
        except (Exception, ValueError, AttributeError):
            pass

        if spreadsheet_reference_date is None:
            raise Exception(
                'A data de referência não foi informada ou não está no formato dd/mm/aaaa.'
            )
        elif spreadsheet_reference_hour is None:
            raise Exception(
                'A hora de referência não foi informada ou não está no formato hh:mm:ss'
            )
        elif spreadsheet_offset_hour is None:
            raise Exception('A diferença de horas deve ser informada.')
        elif spreadsheet_offset is None:
            raise Exception('A diferença de dias deve ser informada.')
        elif spreadsheet_label_reference_date is None:
            raise Exception('A variável de data de referência deve ser informada.')

        spreadsheet_send_time = datetime_string_parser(
            date_str=raw_flow[reader.FlowColumns.SEND_TIME],
            pattern=TIME_FORMAT
        )

        campaign = self.campaign_service.create_campaign(
            name=spreadsheet_campaign_name,
            group=group,
            org=org,
            user=user
        )

        config = self.campaign_service.set_config_campaign_event(
            offset=spreadsheet_offset,
            offset_hour=spreadsheet_offset_hour,
            send_time=spreadsheet_send_time,
            reference_date=spreadsheet_reference_date,
            reference_hour=spreadsheet_reference_hour,
        )

        relative_to = self.contact_service.get_contacfield_by_label(
            org=org,
            label=spreadsheet_label_reference_date
        )

        campaign_event = self.campaign_service.create_campaign_event(  # noqa
            campaign=campaign,
            flow=flow,
            relative_to=relative_to,
            configuration=config
        )

        return campaign, campaign_event, config

    def fetch_group_name_in_spreadsheet(
        self, raw_flow: dict, reader: FlowGoogleSheets,
        organization: RapidProOrganization
    ):
        spreadsheet_group = raw_flow[reader.FlowColumns.CONTACT_GROUP]

        group = self.contact_service.get_contact_group_by_name(
            name=spreadsheet_group,
            org_id=organization.id
        )

        return group

    @ensure_transaction
    def create_sheduled_flow(
        self,
        link: str,
        page: str,
        end_selection: str = None
    ):

        reader = self.spreadsheet_reader(link, page)
        registers = reader.read_sheet_data()

        registers = self.prepare_flow_copy_spreadsheet(registers, reader)
        registers = break_list_into_groups(registers, chunk_size=2)

        with self.db_connection.session.begin():
            user = self.user_service.get_default_user()

            list_serialized_data = []
            spreadsheet_pos = []
            for raw_flow, template in registers:
                spreadsheet_pos.append(
                    {'flow': raw_flow['index'], 'template': template['index']}
                )

                flow, revision, org_destination = self.clone_flow(
                    raw_flow=raw_flow,
                    template=template,
                    user=user,
                    reader=reader
                )

                group = self.fetch_group_name_in_spreadsheet(
                    raw_flow=raw_flow,
                    reader=reader,
                    organization=org_destination
                )

                campaign, campaign_event, config_time = self.create_campaign(
                    reader=reader,
                    raw_flow=raw_flow,
                    flow=flow,
                    org=org_destination,
                    user=user,
                    group=group
                )

                serialized_data = [
                    serialize(
                        serializer_class=self.OrganizationSerializer,
                        instance=org_destination
                    ),
                    serialize(
                        serializer_class=self.FlowSerializer,
                        instance=flow
                    ),
                    serialize(
                        serializer_class=self.FlowRevisionSerializer,
                        instance=revision
                    ),
                    serialize(
                        serializer_class=self.ContactGroupSerializer,
                        instance=group
                    ),
                    serialize(
                        serializer_class=self.CampaignSerializer,
                        instance=campaign
                    ),
                    serialize(
                        serializer_class=self.CampaignEventSerializer,
                        instance=campaign_event,
                        config_time=config_time,
                    )
                ]

                keys = [
                    'organization', 'flow', 'flow_revision', 'group',
                    'campaign', 'campaign_event'
                ]

                list_serialized_data.append(
                    json_representation(keys, serialized_data)
                )

            for data in list_serialized_data:
                notificator = SlackFlowSchedulledNotification(
                    title='Fluxo Criado e Agendado!',
                    data=data
                )
                notificator.send_notification()

            reader.write_sheet_data(
                rows=spreadsheet_pos,
                value=reader.FlowColumns.CONTROL_TO_UPLOAD,
                control_column=reader.FlowColumns.UPLOAD
            )

            return list_serialized_data



from typing import List

from movva_tools.services.base_service import BaseService
from movva_tools.exceptions import (
    ObjectDoesNotExistException,
)

from movva_tools.models.channel_models import RapidProChannel
from movva_tools.models.organization_models import RapidProOrganization


class ChannelService(BaseService):

    def __init__(self, db_connection=None) -> None:
        super().__init__(db_connection)

        # model table entities
        self.Channel = RapidProChannel

    def get_channel_by_name(self, name: str, org: RapidProOrganization):
        channel = self.db_connection.session.query(
            self.Channel
        ).filter_by(
            name=name,
            org_id=org.id
        ).first()

        if channel:
            return channel
        else:
            raise ObjectDoesNotExistException(dababase_object=self.Channel)

    def get_channel_by_id(self, id: int):
        channel = self.db_connection.session.query(
            self.Channel
        ).filter_by(
            id=id,
        ).first()

        if channel:
            return channel
        else:
            raise ObjectDoesNotExistException(dababase_object=self.Channel)

    def get_channel_by_address(self, address: str, org: RapidProOrganization):
        channel = self.db_connection.session.query(
            self.Channel
        ).filter_by(
            address=address,
            org_id=org.id
        ).first()

        if channel:
            return channel
        else:
            raise ObjectDoesNotExistException(dababase_object=self.Channel)



from movva_tools.exceptions import ObjectDoesNotExistException
from movva_tools.models.messages_models import RapidProMessages
from movva_tools.services.base_service import BaseService


class MessageService(BaseService):
    def __init__(self, db_connection=None) -> None:

        super().__init__(db_connection=db_connection)

        # model table entities
        self.Message = RapidProMessages

    def fetch_message_by_id(self, message_id):
        message = self.db_connection.session.query(self.Message).filter_by(
            id=message_id
        ).first()

        if message:
            return message
        else:
            raise ObjectDoesNotExistException(dababase_object=self.Message)

    def fetch_messages_by_channel_and_org(self, channel_id, org_id, direction, status=None):
        if status:
            messages = self.db_connection.session.query(self.Message).filter_by(
                status=status,
                direction=direction,
                channel_id=channel_id,
                org_id=org_id
            ).all()
        else:
            messages = self.db_connection.session.query(self.Message).filter_by(
                channel_id=channel_id,
                direction=direction,
                org_id=org_id
            ).all()

        return messages if messages else []




from movva_tools.services.flow_service import FlowCopyService
from movva_tools.integrations.google_integration import GoogleSheetsIntegration
import pandas as pd


def flow_copy_with_spreadsheets_whatsapp():
    integration = GoogleSheetsIntegration(
        link='spreadsheet_link',
        spreadsheet_page='Whatsapp',
        creds_path='credentials.json'
    )

    data = integration.read_sheet_data()

    # variável para auxiliar a colocar os parâmetros no serviço (opcional)
    df_aux = pd.DataFrame(data[1:], columns=data[0])

    base_flow_name = df_aux['FLUXO ORIGINAL'][0]
    flow_new_name = df_aux['NOVO FLUXO'][0]
    destination_organization_name = df_aux['ORGANIZACAO DESTINO'][0]
    origin_organization_name = df_aux['ORGANIZACAO ORIGEM'][0]

    service = FlowCopyService()

    service.clone(
        base_flow_name=base_flow_name,
        flow_suggested_name=flow_new_name,
        destiny_organization_name=destination_organization_name,
        origin_organization_name=origin_organization_name,
        data=data
    )


def flow_copy_with_spreadsheets_sms():
    integration = GoogleSheetsIntegration(
        link='spreadsheet_link',
        spreadsheet_page='SMS',
        creds_path='credentials.json'
    )

    data = integration.read_sheet_data()

    # variável para auxiliar a colocar os parâmetros no serviço (opcional)
    df_aux = pd.DataFrame(data[1:], columns=data[0])

    base_flow_name = df_aux['FLUXO ORIGINAL'][0]
    flow_new_name = df_aux['NOVO FLUXO'][0]
    destination_organization_name = df_aux['ORGANIZACAO DESTINO'][0]
    origin_organization_name = df_aux['ORGANIZACAO ORIGEM'][0]

    service = FlowCopyService()

    service.clone(
        base_flow_name=base_flow_name,
        flow_suggested_name=flow_new_name,
        destiny_organization_name=destination_organization_name,
        origin_organization_name=origin_organization_name,
        data=data
    )


if __name__ == "__main__":
    flow_copy_with_spreadsheets_whatsapp()



class FlowRevisionParser:
    def __init__(self, flow_json: dict):
        self.flow_json = flow_json

    def extract_all_nodes(self, node=None):
        if node is None:
            node = self.flow_json
        nodes = []
        nodes.append(node)
        if 'nodes' in node:
            for child_node in node['nodes']:
                nodes.extend(self.extract_all_nodes(child_node))
        return nodes

    def extract_texts_from_nodes(self, nodes):
        texts = []
        for node in nodes:
            if 'text' in node:
                texts.append(node['text'])
        return texts

    def extract_channels_from_nodes(self, nodes):
        channels = []
        for node in nodes:
            if 'channel' in node:
                channels.append(node['channel'])
        return channels

    def parse_flow_revision(self):
        all_nodes = self.extract_all_nodes()
        all_texts = self.extract_texts_from_nodes(all_nodes)
        all_channels = self.extract_channels_from_nodes(all_nodes)
        return all_texts, all_channels


# Exemplo de uso
base_flow_json = { ... }  # Substitua por seu JSON de revisão de fluxo
parser = FlowRevisionParser(base_flow_json)
all_texts, all_channels = parser.parse_flow_revision()

for i in range(len(all_texts)):
    print("Text:", all_texts[i])
    print("Channel:", all_channels[i])
    print()



from movva_tools.models.api_token_model import RapidProApiToken
from movva_tools.models.campaign_models import RapidProCampaigns, RapidProCampaignEvents
from movva_tools.models.flow_models import (
    RapidProFlows, RapidProFlowLabel,
    RapidProFlowGroupDependencies,
    RapidProFlowFieldDependencies,
    RapidProFlowsFlowLabels,
    RapidProFlowsRevision,
    RapidProFlowChannelDependencies,
    RapidProFlowFlowLabelDependencies
)
from movva_tools.models.organization_models import RapidProOrganization
from movva_tools.models.channel_models import RapidProChannel, RapidProChannelLog
from movva_tools.models.contacts_models import (
    RapidProContactFields, RapidProContactGroups,
    RapidProContactGroupsContacts
)
from movva_tools.models.user_models import RapidProUser
from movva_tools.models.messages_models import RapidProMessages
